module ImageCTFComparisons
    use Images
    use ContrastTransferFunctions
    type ImageCTFComparison
        type(Image)                     ::  img
        type(ContrastTransferFunction)  ::  ctf
        real                            ::  pixel_size  !<  Angstroms
    end type

    contains

    !>  \brief  Scoring function to be minimised
    function ctffind_objective_function(array_of_values,scoring_parameters) bind(c)
        use iso_c_binding
        use Units
        use ContrastTransferFunctions
        ! arguments
        type(c_ptr), value ::  array_of_values
        type(c_ptr), value ::  scoring_parameters
        ! result
        real(kind=c_double) ::  ctffind_objective_function
        ! private variables
        real(kind=8),               pointer ::  values_to_try(:)
        type(ImageCTFComparison),   pointer ::  comparison_object
        type(ContrastTransferFunction)      ::  my_ctf
        logical,                parameter   ::  debug = .false.
        ! start work
        call c_f_pointer(array_of_values,values_to_try,[3])
        call c_f_pointer(scoring_parameters,comparison_object)
        ! Set defocus values & astigmatism
        my_ctf = comparison_object%ctf
        call my_ctf%SetDefocus( convert(real(values_to_try(1)),angstroms,pixels,comparison_object%pixel_size), &
                                convert(real(values_to_try(2)),angstroms,pixels,comparison_object%pixel_size), &
                                convert(real(values_to_try(3)),degrees,radians) )
        ctffind_objective_function = - comparison_object%img%GetCorrelationWithCTF(my_ctf)
        if (debug) then
            write(*,'(a,3(f0.2,1x),f0.3)') '**debug(ctffind_objective_function): values, score: ', &
                                            values_to_try,ctffind_objective_function
        endif
    end function ctffind_objective_function

end module ImageCTFComparisons

program ctffind

    use Globals
    use ImageFiles
    use Images
    use BruteForceSearches

    implicit none

    call main()

    contains

    subroutine main()

        use UserInputs
        use UserSuppliedParameters
        use EmpiricalDistributions
        use Histograms
        use ContrastTransferFunctions
        use ImageCTFComparisons
        use BruteForceSearches
        use ConjugateGradients
        use Units
        use StringManipulations
        use UsefulFunctions
        use NumericTextFiles
        use DatesAndTimes
        use ProgressBars
        use omp_lib
        use, intrinsic :: iso_c_binding

        ! User input object
        type(UserInput)                 ::  my_user_input

        ! The user will supply values for these
        type(UserSuppliedFilename)      ::  input_filename
        type(UserSuppliedLogical)       ::  input_is_a_movie
        type(UserSuppliedInteger)       ::  number_of_frames_to_average
        type(UserSuppliedFilename)      ::  output_diagnostic_filename
        type(UserSuppliedReal)          ::  pixel_size
        type(UserSuppliedReal)          ::  acceleration_voltage
        type(UserSuppliedReal)          ::  spherical_aberration
        type(UserSuppliedReal)          ::  amplitude_contrast
        type(UserSuppliedInteger)       ::  box_size
        type(UserSuppliedReal)          ::  minimum_resolution
        type(UserSuppliedReal)          ::  maximum_resolution
        type(UserSuppliedReal)          ::  minimum_defocus
        type(UserSuppliedReal)          ::  maximum_defocus
        type(UserSuppliedReal)          ::  defocus_search_step
        type(UserSuppliedReal)          ::  astigmatism_tolerance

        ! Variables
        type(ImageFile)                 ::  input_file
        type(Image)                     ::  current_tile
        type(Image)                     ::  current_input_image
        type(Image)                     ::  current_power_spectrum
        type(Image)                     ::  resampled_power_spectrum
        type(Image)                     ::  average_spectrum
        type(Image)                     ::  sum_spectrum
        type(Image)                     ::  temp_image
        type(Image)                     ::  current_input_image_square
        integer                         ::  current_micrograph_number
        integer                         ::  number_of_micrographs
        integer                         ::  number_of_movie_frames
        integer                         ::  current_input_location
        integer                         ::  current_output_location
        integer                         ::  address_of_tile_center(3)
        type(EmpiricalDistribution)     ::  distribution_of_tile_rms
        type(Histogram)                 ::  histogram_of_tile_rms
        real                            ::  min_rms, max_rms
        integer                         ::  current_tile_number_x, current_tile_number_y,current_tile_number_z
        real                            ::  rms_of_current_tile
        integer                         ::  current_iteration
        integer                         ::  number_of_tiles_used
        integer                         ::  convolution_box_size
        real                            ::  distance_between_zeroes
        type(ContrastTransferFunction)  ::  current_ctf
        type(ImageCTFComparison), target::  comparison_object
        type(BruteForceSearch)          ::  brute_force_search
        real(kind=8)                    ::  bf_midpoint(3), bf_halfrange(3), bf_stepsize(3)
        real(kind=8),   allocatable     ::  cg_starting_point(:)
        real(kind=8),   allocatable     ::  cg_accuracy(:)
        type(ConjugateGradient)         ::  conjugate_gradient
        real                            ::  threshold
        real                            ::  estimated_astigmatism_angle
        type(c_ptr), allocatable        ::  bf_param_array(:)
        character(len=filename_max_len) ::  output_text_fn
        character(len=512)              ::  comment_lines(4)
        character(len=5)                ::  program_version
        type(NumericTextFile)           ::  output_text
        type(NumericTextFile)           ::  output_text_avrot
        type(DateAndTime)               ::  my_date_and_time
        integer,    parameter           ::  overtiling_factor   =  2
        logical,    parameter           ::  whole_micrograph_ft = .true.
        logical,    parameter           ::  replace_bad_tiles_with_plane = .false. ! before doing whole-micrograph FT
        logical,    parameter           ::  keep_all_tiles = .false.
        logical,    parameter           ::  taper_micrograph = .true.
        logical,    parameter           ::  find_astigmatism_angle_using_mirror = .true. ! this saves time, compared to searching for the angle at the same time as the defocus values
        logical,    parameter           ::  do_second_refinement_with_less_low_frequencies = .false.
        real,       parameter           ::  height_of_tails_to_remove = 0.0    ! Tails of the distribution of RMS values of tiles. Reasonable value ~ 0.05 - 0.1
        !integer,    parameter           ::  default_number_of_frames_to_average = 4   ! Average frames in real space before any processing is done
        !integer                         ::  number_of_frames_to_average
        integer                         ::  current_frame_within_average
        real                            ::  values_to_write_out(6)
        real(kind=8),   allocatable     ::  spatial_frequency(:)
        real(kind=8),   allocatable     ::  rotational_average(:), rotational_average_astig(:), rotational_average_astig_fit(:)
        real(kind=8),   allocatable     ::  frc_of_fit(:), frc_of_fit_sigma(:)
        integer                         ::  first_iteration_to_run
        integer                         ::  micrograph_square_dimension
        type(ProgressBar)               ::  my_progress_bar
        integer                         ::  counter


        ! Variables specifically for the old-school input
        logical                         ::  old_school_input
        character(len=200)              ::  temp_string
        integer                         ::  arg_status
        real                            ::  xmag, dstep
        integer                         ::  current_arg


        !!
        !! Start
        !!


        ! Initialise the program
        program_version = '4.0.7'
        call this_program%Init('CTFFind',program_version,'2014')

        ! Check whether we should expect old-school input
        old_school_input = .false.
        do current_arg=1,command_argument_count()
            call get_command_argument(current_arg,temp_string,status=arg_status)
            if (arg_status .eq. 0 .and. .not. old_school_input) then
                old_school_input = StringsAreEqual(temp_string,'--old-school-input')
            endif
        enddo


        ! Interact with the user
        call my_user_input%Init(this_program%program_name)
        if (old_school_input) then
            !
            ! The old way of interacting with users. To be deprecated.
            !
            read(*,'(a)')                           temp_string
                                                    input_filename%value = temp_string
            read(*,'(a)')                           temp_string
                                                    output_diagnostic_filename%value = temp_string
            read(*,*)   spherical_aberration%value, acceleration_voltage%value, &
                                                    amplitude_contrast%value, xmag, dstep
            read(*,*)      box_size%value, minimum_resolution%value, maximum_resolution%value, &
                                                    minimum_defocus%value, maximum_defocus%value, defocus_search_step%value, &
                                                    astigmatism_tolerance%value


            pixel_size%value = dstep * 10000.0 / xmag

            input_filename%set_by_user = .true.
            output_diagnostic_filename%set_by_user = .true.
            spherical_aberration%set_by_user = .true.
            acceleration_voltage%set_by_user = .true.
            amplitude_contrast%set_by_user = .true.
            pixel_size%set_by_user = .true.
            box_size%set_by_user = .true.
            minimum_resolution%set_by_user = .true.
            maximum_resolution%set_by_user = .true.
            minimum_defocus%set_by_user = .true.
            maximum_defocus%set_by_user = .true.
            defocus_search_step%set_by_user = .true.
            astigmatism_tolerance%set_by_user = .true.
            number_of_frames_to_average%set_by_user = .false.
            number_of_frames_to_average%value = 1

            ! Input file is normally initialised during the user interaction to work out whether it's a stack of movie frames
            input_is_a_movie%value  =   .false.
            call input_file%Init(input_filename%value)
            if (input_file%GetStackSize() .gt. 1) then
                call this_program%TerminateWithFatalError('ctffind','Input stacks are only supported when using the modern UI')
            endif

            ! Output for old-school users
            write(*,'(/a)') ' CS[mm], HT[kV], AmpCnst, XMAG, DStep[um]'
            write(*,'(f5.1,f9.1,f8.2,f10.1,f9.3/)') spherical_aberration%value, acceleration_voltage%value, amplitude_contrast%value, xmag, dstep

        else
            !
            ! The normal way of interacting with users
            !

        input_filename              =   my_user_input%GetFilenameFromUser(  'Input image file name','Filename of input image',&
                                                                            'input_filename','input.mrc',.true.)
        ! Open the input file
        call input_file%Init(input_filename%value)
        if (input_file%GetStackSize() .gt. 1) then
            input_is_a_movie        =   my_user_input%GetLogicalFromUser(   'Input is a movie (stack of frames)', &
                                                                            'Answer yes if the input file is a stack of '//&
                                                                            'frames from a dose-fractionated movie. If not, '//&
                                                                            'each image will be processed separately', &
                                                                            'input_is_a_movie','no')
        else
            input_is_a_movie%value  =   .false.
        endif

        if (input_is_a_movie%value) then
            number_of_frames_to_average = my_user_input%GetIntegerFromUser( 'Number of frames to average together', &
                                                                            'If the number of electrons per frame is too low, '//&
                                                                            'there may be strong artefacts in the '//&
                                                                            'estimated power spectrum. This can be '//&
                                                                            'alleviated by averaging frames '//&
                                                                            'with each other in real space '//&
                                                                            'before computing Fourier transforms.', &
                                                                            'num_frames_to_average','1')
        else
            number_of_frames_to_average%value = 1
        endif

        ! Continue with user interaction
        output_diagnostic_filename  =   my_user_input%GetFilenameFromUser(  'Output diagnostic filename', &
                                                                            'Will contain experimental power spectrum '//&
                                                                            'and best fit', &
                                                                            'output_diag_filename','diagnostic_output.mrc',.false.)

        pixel_size                  =   my_user_input%GetRealFromUser(      'Pixel size','in Angstroms','pixel_size', &
                                                                            '1.0',min_value=0.0)

        acceleration_voltage        =   my_user_input%GetRealFromUser(      'Acceleration voltage', &
                                                                            'in kV','acceleration_voltage','300.0',min_value=0.0)
        spherical_aberration        =   my_user_input%GetRealFromUser(      'Spherical aberration', &
                                                                            'in mm','spherical_aberration','2.7',min_value=0.0)
        amplitude_contrast          =   my_user_input%GetRealFromUser(      'Amplitude contrast', &
                                                                            'fraction of amplitude contrast','amplitude_contrast', &
                                                                            '0.07',min_value=0.0,max_value=1.0)
        box_size                    =   my_user_input%GetIntegerFromUser(   'Size of power spectrum to compute', &
                                                                            'Size of the power spectrum (pixels) to compute', &
                                                                            'box_size','512',min_value=0)
        minimum_resolution          =   my_user_input%GetRealFromUser(      'Minimum resolution', &
                                                                            'Lowest resolution used for fitting CTF (Angstroms)', &
                                                                            'minimum_resolution','30.0', &
                                                                            min_value=0.0,max_value=50.0)
        maximum_resolution          =   my_user_input%GetRealFromUser(      'Maximum resolution', &
                                                                            'Highest resolution used for fitting CTF (Angstroms)', &
                                                                            'maximum_resolution','5.0',min_value=0.0)
        minimum_defocus             =   my_user_input%GetRealFromUser(      'Minimum defocus', &
                                                                            'Positive values for underfocus. Lowest value to '//&
                                                                            'search over. (Angstroms)','minimum_defocus', &
                                                                            '5000.0',min_value=0.0)
        maximum_defocus             =   my_user_input%GetRealFromUser(      'Maximum defocus', &
                                                                            'Positive values for underfocus. Highest value to '//&
                                                                            'search over. (Angstroms)','maximum_defocus', &
                                                                            '50000.0',min_value=0.0)
        defocus_search_step         =   my_user_input%GetRealFromUser(      'Defocus search step', &
                                                                            'Step size for defocus search (Angstroms)', &
                                                                            'defocus_search_step','500.0',min_value=1.0)
        astigmatism_tolerance       =   my_user_input%GetRealFromUser(      'Expected (tolerated) astigmatism', &
                                                                            'Astigmatism values much larger than this will be '//&
                                                                            'penalised (Angstroms; set to 0.0 to remove this '//&
                                                                            'restraint)', &
                                                                            'astigmatism_tolerance','100.0',min_value=0.0)

        endif ! end of check for old-school input

        ! Done interacting with the user
        call my_user_input%UpdateDefaults()

        ! We want to time the runtime, not the user interaction
        call this_program%SetStartDate()

        ! Do some argument checking
        if (minimum_resolution%value .lt. maximum_resolution%value) then
            call this_program%TerminateWithFatalError('ctffind','Minimum resolution must be greater than maximum resolution')
        endif
        if (minimum_defocus%value .gt. maximum_defocus%value) then
            call this_program%TerminateWithFatalError('ctffind','Minimum defocus must less than maximum defocus')
        endif

        ! How many micrographs are we dealing with?
        if (input_is_a_movie%value) then
            ! we only support 1 movie per file
            number_of_movie_frames = input_file%GetStackSize()
            number_of_micrographs  = 1
        else
            number_of_movie_frames = 1
            number_of_micrographs  = input_file%GetStackSize()
        endif


        ! Print out information about input file
        call input_file%PrintInfo()

        ! Check micrograph is square if it needs to be
        if (whole_micrograph_ft) then
            if (input_file%GetDimension(1) .ne. input_file%GetDimension(2)) then
                !call this_program%TerminateWithFatalError('ctffind','micrograph must be square')
            endif
        endif

        ! Prepare the tile image
        call current_tile%Allocate(dims=[box_size%value,box_size%value,1])
        call average_spectrum%Allocate(current_tile,in_real_space=.true.)

        ! Testing
        !call current_ctf%Init(300.0,2.3,0.07,1.2,3.9,30.0,1.0/30.0,1.0/3.0,100.0,2.08)
        !call average_spectrum%ComputeRotationalAverageOfPowerSpectrum(current_ctf,rotational_average_astig,rotational_average_astig_fit)
        !call this_program%TerminateWithFatalError('dbg','dbg')

        ! Prepare the output text file
        output_text_fn = output_diagnostic_filename%value
        call FilenameReplaceExtension(output_text_fn,'txt')
        call output_text%Init(output_text_fn,OPEN_TO_WRITE,6)
        write(comment_lines(1),'(4a)') 'Output from CTFFind version ', program_version, &
                                            ', run on ', my_date_and_time%DateAndTimeAsString()
        write(comment_lines(2),'(3a,i0)') 'Input file: ', input_filename%value, '; Number of micrographs: ', number_of_micrographs
        write(comment_lines(3),'(a,f0.3,a,f0.1,a,f0.1,a,f0.2)') 'Pixel size: ', pixel_size%value, &
                    '; acceleration voltage: ', acceleration_voltage%value, &
                    '; spherical aberration: ', spherical_aberration%value, &
                    '; amplitude contrast: ', amplitude_contrast%value
        write(comment_lines(4),'(a,i0,4(a,f0.1))') 'Box size: ', box_size%value, &
                    '; min res: ', minimum_resolution%value, '; max. res.: ', maximum_resolution%value, &
                    '; min. def.: ', minimum_defocus%value, '; max. def.: ', maximum_defocus%value

        call output_text%WriteCommentLine(comment_lines(1))
        call output_text%WriteCommentLine(comment_lines(2))
        call output_text%WriteCommentLine(comment_lines(3))
        call output_text%WriteCommentLine(comment_lines(4))
        call output_text%WriteCommentLine('Columns: #1 - micrograph number; #2 - defocus 1; #3 - defocus 2; '//&
                                          '#4 - azimuth of astigmatism; #5 - cross correlation; #6 - spacing '//&
                                          '(in Angstroms) up to which CTF rings were detected')

        ! Prepare a text file with 1D rotational average spectra
        call FilenameAddSuffix(output_text_fn,'_avrot')


        if (.not. old_school_input .and. number_of_micrographs .gt. 1) then
            write(*,'(/a,i0,a)') 'Will estimate the CTF parameters for ', number_of_micrographs, ' micrographs.'
            write(*,'(2a/)')     'Results will be written to this file: ', trim(adjustl(output_text%filename))
            call my_progress_bar%Begin(number_of_micrographs)
        endif


        ! Loop over micrographs
        do current_micrograph_number=1,number_of_micrographs
            if (old_school_input .or. number_of_micrographs .eq. 1) then
                write(*,'(/a,i0,a,i0)') 'Working on micrograph ', current_micrograph_number, ' of ', number_of_micrographs
            endif
            !
            ! We loop twice over the tiles:
            ! Iteration 1: get statistics of tiles. We want to know their RMS values.
            ! Iteration 2: accumulate average power spectrum of selected tiles
            !
            ! If we are using the whole micrograph and not tiling, we skip iteration 1
            if (whole_micrograph_ft) then
                first_iteration_to_run = 2
            else
                first_iteration_to_run = 1
            endif
            number_of_tiles_used = 0
            average_spectrum = 0.0e0
            do current_iteration=first_iteration_to_run,2
                if (current_iteration .eq. 1) call distribution_of_tile_rms%Init(keep_sample_values=.true.)
                do current_tile_number_z=1,number_of_movie_frames,number_of_frames_to_average%value
                    do current_frame_within_average=1,number_of_frames_to_average%value
                        current_input_location = current_tile_number_z &
                                               + number_of_movie_frames * (current_micrograph_number-1) &
                                               + (current_frame_within_average-1)
                        if (current_input_location .gt. number_of_movie_frames*number_of_micrographs) cycle
                        call current_input_image%ReadFromDisk(input_file,current_input_location)
                        ! let's get reasonable values
                        call current_input_image%ZeroFloatAndNormalise(100.0e0)
                        ! let's pad to make it square if we need to
                        micrograph_square_dimension = maxval(current_input_image%GetLogicalDimensions())
                        if (Isodd(micrograph_square_dimension)) micrograph_square_dimension = micrograph_square_dimension + 1 ! even may be nicer for fftw
                        call current_input_image_square%Allocate(dims=[micrograph_square_dimension, &
                                                                       micrograph_square_dimension,1])
                        call current_input_image%ClipInto(current_input_image_square, &
                                                          padding_value_real=current_input_image%GetAverageOfValues())
                        current_input_image = current_input_image_square
                        call current_input_image_square%Deallocate()
                        !
                        if (current_frame_within_average .eq. 1) then
                            call temp_image%Allocate(mould=current_input_image)
                            temp_image = 0.0e0
                        endif
                        call temp_image%AddImage(current_input_image)
                    enddo
                    current_input_image = temp_image
                    if (temp_image%IsAllocated()) call temp_image%Deallocate()

                    if (.not. whole_micrograph_ft .or. replace_bad_tiles_with_plane) then
                        do current_tile_number_y=1,int(real(input_file%GetDimension(2))/box_size%value)*overtiling_factor
                            do current_tile_number_x=1,int(real(input_file%GetDimension(1))/box_size%value)*overtiling_factor
                                ! Cut out the tile
                                address_of_tile_center = [(current_tile_number_x-1)*current_tile%GetLogicalDimension(1)    &
                                                                                                    /overtiling_factor+    &
                                                           current_tile%GetPhysicalIndexOfBoxCenter(1),                    &
                                                          (current_tile_number_y-1)*current_tile%GetLogicalDimension(2)    &
                                                                                                    /overtiling_factor+    &
                                                          current_tile%GetPhysicalIndexOfBoxCenter(2),                     &
                                                          1]
                                address_of_tile_center = min(address_of_tile_center,current_input_image%GetLogicalDimensions() &
                                                            - current_tile%GetPhysicalAddressOfBoxCenter())
                                address_of_tile_center(3) = 1
                                call current_input_image%ClipInto(current_tile,address_of_tile_center)
                                call current_tile%RemoveLinearGradient()
                                rms_of_current_tile = current_tile%GetSigmaOfValues()
                                if (current_iteration .eq. 1) call distribution_of_tile_rms%AddSampleValue(rms_of_current_tile)
                                if (current_iteration .eq. 2) then
                                    if (.not. whole_micrograph_ft) then
                                        ! Now we accumulate the average power spectrum
                                        if (      (rms_of_current_tile .lt. max_rms .and. rms_of_current_tile .gt. min_rms) &
                                             .or. keep_all_tiles) then
                                            number_of_tiles_used = number_of_tiles_used + 1
                                            ! Average amplitudes
                                            call current_tile%ComputeAmplitudeSpectrum(current_power_spectrum, &
                                                                                       compute_square=.true.)
                                            call average_spectrum%AddImage(current_power_spectrum)
                                        endif ! tile rms ok
                                    else if (replace_bad_tiles_with_plane) then
                                        if (rms_of_current_tile .ge. max_rms .or. rms_of_current_tile .le. min_rms) then
                                            ! We need to compute the plane and stick it back into the micrograph
                                            call current_input_image%ClipInto(current_tile,address_of_tile_center)
                                            call current_tile%ReplaceWithLinearGradient()
                                            call current_tile%AddInto(current_input_image,address_of_tile_center,overwrite=.true.)
                                        endif
                                    endif ! whole micrograph / replace bad tiles
                                endif ! iteration 2
                            enddo ! tile X
                        enddo ! tile Y
                    else

                    endif ! do_tiling

                    if (whole_micrograph_ft .and. current_iteration .eq. 2) then
                        ! we don't do tiling
                        if (taper_micrograph) then
                            call current_input_image%TaperEdges()
                        endif
                        number_of_tiles_used = number_of_tiles_used + 1
                        call current_input_image%ComputeAmplitudeSpectrum(current_power_spectrum,compute_square=.true., &
                                                                          set_central_pixel_to_zero=.true.)
                        ! Use the lines below if averaging of tile PS should be done before downsampling
                        !call sum_spectrum%AddImage(current_power_spectrum)

                        ! Use the lines below if downsampling should be done before averaging of tile PS
                        call current_power_spectrum%ForwardFFT()
                        call resampled_power_spectrum%Allocate(dims=[box_size%value,box_size%value,1],in_real_space=.false.)
                        call current_power_spectrum%ClipInto(resampled_power_spectrum)
                        call resampled_power_spectrum%BackwardFFT()

                        call average_spectrum%AddImage(resampled_power_spectrum)

                    endif


                enddo ! movie frames (tile Z)

                if (current_iteration .eq. 1) then
                    if (.not. whole_micrograph_ft .or. replace_bad_tiles_with_plane) then
                        call distribution_of_tile_rms%PopulateHistogram(histogram_of_tile_rms,100)
                        ! We will only process tiles which have RMS values in this range
                        min_rms = histogram_of_tile_rms%FirstValueWithCountAboveFractionOfMaxCount(height_of_tails_to_remove)
                        !min_rms = min_rms + histogram_of_tile_rms%GetBinWidth() ! This is a quirk of CTFFIND
                        max_rms = histogram_of_tile_rms%LastValueWithCountAboveFractionOfMaxCount(height_of_tails_to_remove)
                    endif
                else
                    if (whole_micrograph_ft) then
                        !! we don't do tiling
                        !if (taper_micrograph) then
                        !    !call current_input_image%WriteToDisk('dbg_before_taper.mrc')
                        !    call current_input_image%TaperEdges()
                        !endif
                        !number_of_tiles_used = number_of_tiles_used + 1
                        !!call current_input_image%WriteToDisk('dbg_before_ft.mrc')
                        !call current_input_image%ComputeAmplitudeSpectrum(current_power_spectrum,compute_square=.true.)
                        !call current_power_spectrum%ForwardFFT()
                        !call current_tile%Allocate(dims=[box_size%value,box_size%value,1],in_real_space=.false.)
                        !call current_power_spectrum%ClipInto(current_tile)
                        !call current_tile%BackwardFFT()
                        !call average_spectrum%AddImage(current_tile)

                        ! resample the average power spectrum
                        !call sum_spectrum%ForwardFFT()
                        !call current_tile%Allocate(dims=[box_size%value,box_size%value,1],in_real_space=.false.)
                        !call sum_spectrum%ClipInto(average_spectrum)
                        !call average_spectrum%BackwardFFT()
                    else
                        if (number_of_tiles_used .eq. 0) then
                            call this_program%TerminateWithFatalError('ctffind','no tiles survived RMS selection')
                        endif
                        !write(*,'(a,i0,a,i0,a)') '**info(ctffind): ', number_of_tiles_used, ' tiles out of ', &
                        !                ceiling(real(input_file%GetDimension(1))/box_size%value) *  &
                        !                ceiling(real(input_file%GetDimension(2))/box_size%value), &
                        !                ' were used to compute the average power spectrum'
                    endif
                    call average_spectrum%MultiplyByConstant(1.0e0/real(number_of_tiles_used))



                    ! Initialise the CTF object
                    call current_ctf%Init(  acceleration_voltage%value,spherical_aberration%value,amplitude_contrast%value, &
                                            minimum_defocus%value,minimum_defocus%value,0.0e0, &
                                            1.0e0/minimum_resolution%value,1.0e0/maximum_resolution%value, &
                                            astigmatism_tolerance%value,pixel_size%value)
                    call current_ctf%SetDefocus(minimum_defocus%value/pixel_size%value,minimum_defocus%value/pixel_size%value,0.0e0)



                    ! Filter averaged power spectrum
                    !convolution_box_size = int( average_spectrum%GetLogicalDimension(1) &
                    !                           * pixel_size%value &
                    !                           / minimum_resolution%value &
                    !                           * sqrt(2.0e0))

                    distance_between_zeroes = current_ctf%ComputeFrequencyOfAZero(0.0,3)-current_ctf%ComputeFrequencyOfAZero(0.0,2)
                    convolution_box_size = distance_between_zeroes * average_spectrum%GetLogicalDimension(1) * 3.0
                    if (IsEven(convolution_box_size)) convolution_box_size = convolution_box_size + 1
                    !call average_spectrum%WriteToDisk('dbg_average_spectrum_before_conv.mrc')
                    call average_spectrum%SpectrumBoxConvolution(convolution_box_size,current_power_spectrum, &
                                                                 minimum_radius=average_spectrum%logical_dimensions(1) &
                                                                                *pixel_size%value/minimum_resolution%value)

                    !call current_power_spectrum%WriteToDisk('dbg_background.mrc')

                    ! Subtract filtered power spectrum from power spectrum. This should remove background slope.
                    average_spectrum%real_values(1:average_spectrum%GetLogicalDimension(1),:,:) &
                    =   average_spectrum%real_values(1:average_spectrum%GetLogicalDimension(1),:,:)**2 &
                    -   current_power_spectrum%real_values(1:average_spectrum%GetLogicalDimension(1),:,:)**2

                    !call average_spectrum%WriteToDisk('dbg_average_spectrum_background_subtracted.mrc')


                    average_spectrum%real_values( average_spectrum%GetPhysicalIndexOfBoxCenter(1), &
                                                  average_spectrum%GetPhysicalIndexOfBoxCenter(2), &
                                                  average_spectrum%GetPhysicalIndexOfBoxCenter(3)) = 0.0e0

                    ! Threshold high values
                    threshold = average_spectrum%GetMaximumValue(minimum_distance_from_center=3,minimum_distance_from_edge=3) ! used to be 3
                    call average_spectrum%SetMaximumValue(threshold)





                    ! Set up the comparison object
                    comparison_object%ctf = current_ctf
                    comparison_object%img = average_spectrum
                    comparison_object%pixel_size = pixel_size%value

                    ! Do a brute-force search for the CTF parameters
                    if (old_school_input) then
                        write(*,'(/a)') ' SEARCHING CTF PARAMETERS...'
                    endif
                    if (find_astigmatism_angle_using_mirror) then
                        ! First let's look for the astigmatism angle
                        temp_image = average_spectrum
                        call temp_image%ApplyMirror(2)
                        call average_spectrum%SearchForRotationalAlignmentWith(temp_image,90.0,5.0,estimated_astigmatism_angle, &
                                                                      minimum_radius=pixel_size%value/minimum_resolution%value, &
                                                                      maximum_radius=pixel_size%value/maximum_resolution%value)
                        estimated_astigmatism_angle = estimated_astigmatism_angle * 0.5e0
                        !print *, 'Estimated astigmatism angle = ', estimated_astigmatism_angle

                        ! Now we can look for the defocus values
                        bf_halfrange = [real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                                        real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                                        0.0d0]
                        bf_midpoint  = [real(minimum_defocus%value,kind=8), &
                                        real(minimum_defocus%value,kind=8), &
                                        real(estimated_astigmatism_angle,kind=8)]
                        bf_midpoint  = bf_midpoint + bf_halfrange
                        bf_stepsize  = [real(defocus_search_step%value,kind=8),real(defocus_search_step%value,kind=8), 0.0d0]
                    else

                        bf_halfrange = [real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                                        real(0.5*(maximum_defocus%value-minimum_defocus%value),kind=8), &
                                        45.0d0]
                        bf_midpoint  = [real(minimum_defocus%value,kind=8),real(minimum_defocus%value,kind=8),0.0d0]
                        bf_midpoint  = bf_midpoint + bf_halfrange
                        bf_stepsize  = [real(defocus_search_step%value,kind=8),real(defocus_search_step%value,kind=8), 5.0d0]

                    endif

                    ! Actually run the BF search
#ifdef _OPENMP
                    if (.not. allocated(bf_param_array)) allocate(bf_param_array(omp_get_max_threads()))
#else
                    if (.not. allocated(bf_param_array)) allocate(bf_param_array(1))
#endif
                    bf_param_array = c_loc(comparison_object)
                    call brute_force_search%Init(ctffind_objective_function,bf_param_array,3,    &
                                                 bf_midpoint,bf_halfrange,bf_stepsize, &
                                                 print_progress_bar=number_of_micrographs .eq. 1)
                    call brute_force_search%Run()
                    cg_starting_point = brute_force_search%GetBestValues()

                    ! Remember the results of the search
                    call current_ctf%SetDefocus(real(cg_starting_point(1))/pixel_size%value, &
                                                real(cg_starting_point(2))/pixel_size%value, &
                                                convert(real(cg_starting_point(3)),degrees,radians))

                    ! Print out results of brute force search
                    if (old_school_input) then
                        write(*,'(a)')  '      DFMID1      DFMID2      ANGAST          CC'
                        write(*,'(3(f12.2),f12.5,a)')   current_ctf%GetDefocus1InAngstroms(pixel_size%value), &
                                                        current_ctf%GetDefocus2InAngstroms(pixel_size%value), &
                                                        current_ctf%GetAstigmatismAzimuthInDegrees(), &
                                                        -brute_force_search%GetBestScore(), &
                                                         '  '
                    endif

                    ! Now we refine in the neighbourhood by using Powell's conjugate gradient algorithm
                    if (old_school_input) then
                        write(*,'(/a)') ' REFINING CTF PARAMETERS...'
                        write(*,'(a)')  '      DFMID1      DFMID2      ANGAST          CC'
                    endif
                    cg_accuracy = [100.0d0,100.0d0,0.5d0]
                    call conjugate_gradient%Init(3,ctffind_objective_function,c_loc(comparison_object), &
                                                cg_starting_point,cg_accuracy)
                    call conjugate_gradient%Run()

                    ! Do a second refinement with less of the low frequencies
                    if (do_second_refinement_with_less_low_frequencies) then
                        call comparison_object%ctf%SetLowestFrequencyForFitting(   &
                                    comparison_object%ctf%GetLowestFrequencyForFitting() &
                                  +( comparison_object%ctf%GetHighestFrequencyForFitting() &
                                    -comparison_object%ctf%GetLowestFrequencyForFitting())*0.25)
                        cg_starting_point = conjugate_gradient%GetBestValues()
                        call current_ctf%SetDefocus(real(cg_starting_point(1))/pixel_size%value, &
                                                real(cg_starting_point(2))/pixel_size%value, &
                                                convert(real(cg_starting_point(3)),degrees,radians))
                        if (old_school_input) then
                            write(*,'(3(f12.2),f12.5,a/)')   current_ctf%GetDefocus1InAngstroms(pixel_size%value), &
                                                        current_ctf%GetDefocus2InAngstroms(pixel_size%value), &
                                                        current_ctf%GetAstigmatismAzimuthInDegrees(), &
                                                        -conjugate_gradient%GetBestScore(), &
                                                        '  Final Values'
                        endif
                        call conjugate_gradient%Run()
                    endif


                    ! Remember the results of the refinement
                    cg_starting_point = conjugate_gradient%GetBestValues()
                    call current_ctf%SetDefocus(real(cg_starting_point(1))/pixel_size%value, &
                                                real(cg_starting_point(2))/pixel_size%value, &
                                                convert(real(cg_starting_point(3)),degrees,radians))



                    ! Generate diagnostic image
                    current_output_location = current_micrograph_number
                    call average_spectrum%MultiplyByConstant(1.0e0/average_spectrum%GetSigmaOfValues())
                    call average_spectrum%AddConstant(0.5e0)
                    call average_spectrum%Compute1DRotationalAverage(rotational_average)
                    call average_spectrum%ComputeRotationalAverageOfPowerSpectrum(current_ctf,spatial_frequency, &
                                                                                  rotational_average_astig, &
                                                                                  rotational_average_astig_fit, &
                                                                                  frc_of_fit,frc_of_fit_sigma)
                    call average_spectrum%OverlayCTF(current_ctf)
                    call average_spectrum%WriteToDisk(output_diagnostic_filename%value,current_output_location)

                    ! Until what frequency were CTF rings detected?
                    do counter=2,size(frc_of_fit)
                        if (count(frc_of_fit(1:counter-1) .gt. 0.2d0) .gt. 3 .and. frc_of_fit(counter) .le. 0.2d0) then
                            !write(*,'(a,f0.1,a)') 'CTF oscillations were detected up to ', &
                            !                       pixel_size%value/spatial_frequency(i), ' Angstroms'
                            exit
                        endif
                    enddo

                    ! Print results out to the terminal
                    if (old_school_input) then
                        write(*,'(3(f12.2),f12.5,a/)')   current_ctf%GetDefocus1InAngstroms(pixel_size%value), &
                                                        current_ctf%GetDefocus2InAngstroms(pixel_size%value), &
                                                        current_ctf%GetAstigmatismAzimuthInDegrees(), &
                                                        -conjugate_gradient%GetBestScore(), &
                                                        '  Final Values'
                    else if (number_of_micrographs .eq. 1) then
                        write(*,'(a,f0.2,1x,a,1x,f0.2,a)')  'Estimated defocus values        : ', &
                                                        current_ctf%GetDefocus1InAngstroms(pixel_size%value), ',', &
                                                        current_ctf%GetDefocus2InAngstroms(pixel_size%value), ' Angstroms'
                        write(*,'(a,f0.2,a)')               'Estimated azimuth of astigmatism: ', &
                                                        current_ctf%GetAstigmatismAzimuthInDegrees(), ' degrees'
                        write(*,'(a,f0.5)')                 'Score                           : ', &
                                                        -conjugate_gradient%GetBestScore()
                        write(*,'(a,f0.1,a)')               'Thon rings detectable up to     : ', &
                                                        pixel_size%value/spatial_frequency(counter), ' Angstroms'
                    endif


                    ! Write out results
                    values_to_write_out(1) = real(current_micrograph_number)
                    values_to_write_out(2:4) = current_ctf%GetDefocusParametersInAngstromsAndDegrees(pixel_size%value)
                    values_to_write_out(5) = -conjugate_gradient%GetBestScore()
                    values_to_write_out(6) = pixel_size%value/spatial_frequency(counter)
                    call output_text%WriteDataLine(values_to_write_out)

                    ! Write avrot
                    if (output_text_avrot%number_of_data_lines .eq. 0) then
                        call output_text_avrot%Init(output_text_fn,OPEN_TO_WRITE,size(rotational_average))
                        call output_text_avrot%WriteCommentLine(comment_lines(1))
                        call output_text_avrot%WriteCommentLine(comment_lines(2))
                        call output_text_avrot%WriteCommentLine(comment_lines(3))
                        call output_text_avrot%WriteCommentLine(comment_lines(4))
                        call output_text_avrot%WriteCommentLine('6 lines per micrograph: #1 - spatial frequency (1/pixels); '//&
                                                         '#2 - 1D rotational average of spectrum (assuming no astigmatism); '//&
                                                         '#3 - 1D rotational average of spectrum; #4 - CTF fit; '//&
                                                         '#5 - cross-correlation between spectrum and CTF fit; '//&
                                                         '#6 - 2sigma of expected cross correlation of noise')
                    endif
                    call output_text_avrot%WriteDataLine(real(spatial_frequency))
                    call output_text_avrot%WriteDataLine(real(rotational_average))
                    call output_text_avrot%WriteDataLine(real(rotational_average_astig))
                    call output_text_avrot%WriteDataLine(real(rotational_average_astig_fit))
                    call output_text_avrot%WriteDataLine(real(frc_of_fit))
                    call output_text_avrot%WriteDataLine(real(frc_of_fit_sigma))

                endif
            enddo ! end of loop over the 2 iterations

            ! Mark progress
            if (.not. old_school_input .and. number_of_micrographs .gt. 1) then
                call my_progress_bar%Update(current_micrograph_number)
            endif

        enddo ! end of loop over micrographs

        if (.not. old_school_input .and. number_of_micrographs .gt. 1) then
            call my_progress_bar%Finish()
        endif

        ! Tell the user where the outputs are
        write(*,'(/2a)')'Summary of results                          : ', trim(adjustl(output_text%filename))
        write(*,'(2a)') 'Diagnostic images                           : ', trim(adjustl(output_diagnostic_filename%value))
        write(*,'(2a)') 'Detailled results, including 1D fit profiles: ', trim(adjustl(output_text_avrot%filename))
        write(*,'(2a)') 'Use this command to plot 1D fit profiles    : ctffind_plot_results.sh ', &
                                                                          trim(adjustl(output_text_avrot%filename))



        ! Done running the program
        call this_program%Terminate()

    end subroutine main

end program ctffind
